private import lua;

private import std.math;

private import opengl;

private import type;
private import shot;
private import command;
private import mymath;
private import myio;
private import bl;
private import sound;

class Bullet : public Shot {
public:
	this(float x, float y, float sx, float sy, float a, float s,
		 lua_State lua, char[] fn, Type t, float id) {
		super(x, y, sx, sy, a, s, t);
		command_ = new Command(lua, fn, this);
		wait_ = 0;
		id_ = id;
		bl_ = new BLState();
	}
	this(char[] file) {
		super(0, 0, 0, 0, Type.NONE);
		LuaFile lua = new LuaFile(file);
		command_ = new WaitableCommand(lua, this);
		wait_ = 0;
		bl_ = new BLState();
	}

	void move() {
		if (turn_ == 0) command_.init();
		if (wait_ == 0) {
			command_.run();
		}
		else {
			wait_--;
		}
		super.move();
	}

	void runFunc(char[] fn) {
		command_.run(fn);
	}

	void noticeChangeDirection() {
//		float r = std.math.atan2(sx_, sy_);
//		radian_ = (int)((r / std.math.PI + 1) * 16);
	}

	void wait(int t) {
		wait_ = t;
	}

	float id() { return id_; }

	bit isCounterable() {
		return false;
	}

	void setCounterable() {}

	BLState bl() { return bl_; }

	float prevAngle() { return prevAngle_; }
	void prevAngle(float pa) { prevAngle_ = pa; }

	void addChild(Bullet b) {
		children_ ~= b;
	}
	Bullet[] children() { return children_; }

	void parent(Bullet b) { parent_ = b; }
	Bullet parent() { return parent_; }

protected:
	Command command_;

	int wait_;

	float id_;

	BLState bl_;

	float prevAngle_;

	Bullet[] children_;
	Bullet parent_;
} 

// cannot declare in another source code
class Enemy : Bullet {
public:
	this(float x, float y, float sx, float sy, float a, float s,
		 lua_State lua, char[] fn, float life, Type t, float id)
	in {
		assert(life > 0);
	}
	body {
		super(x, y, sx, sy, a, s, lua, fn, t, id);
		life_ = life;
		score_ = sqrt(cast(float)life) * 100;
		dyingCnt_ = 100000;
		dyingRokFactor_ = 0;
	}

	bit alive() {
		return dyingCnt_ > 0 && life_ > -5000;
	}

	// @@@ property problem?
	float life() {
		return super.life();
	}

	int score() {
		return score_;
	}
	void score(int s) {
		score_ = s;
	}

	void life(float life) {
		super.life(life);
		if (life_ < 0) {
			destroy();
		}
	}

	void move() {
		dyingCnt_--;

		if (dyingCnt_ > 100) {
			super.move();
		}
	}

	int dyingCnt() { return dyingCnt_; }

	void draw() {
		switch (type_) {
		case (-Type.SHOT_CORE):
		case (-Type.ENEMY_WALL):
		case (-Type.ENEMY_BLOCK):
		case (-Type.ENEMY_POINT):
		case (-Type.ENEMY_BALLS):
		{
			static const int EXPLODE_NUM = 3;
			class Explode {
			public:
				int[EXPLODE_NUM] x, y, sx, sy, r, sr;
			}

			if (data_ === null) {
				Explode ex = new Explode();
				for (int j = 0; j < EXPLODE_NUM; j++) {
					ex.x[j] = x_;
					ex.y[j] = y_;
//					ex.sx[j] = rnd(11)-5 + sx_;
//					ex.sy[j] = rnd(11)-5 + sy_;
					ex.sx[j] = rnd(7)-3 + sx_;
					ex.sy[j] = rnd(7)-3 + sy_;
					ex.r[j] = rnd(360);
					ex.sr[j] = rnd(31)-15;
				}
				data_ = ex;
			}

			Explode ex = (Explode)data_;

			int rx = 55-dyingCnt_;
			int ry = rx >> 1;
			glColor4f(1, 1, 1, 0.01*dyingCnt_);

			for (int j = 0; j < EXPLODE_NUM; j++) {
				ex.x[j] += ex.sx[j];
				ex.y[j] += ex.sy[j];
				ex.r[j] += ex.sr[j];

				glPushMatrix();

				glTranslatef(ex.x[j], ex.y[j], 0);
				glRotatef(ex.r[j], 0, 0, 1);

				static const int PNT_NUM = 13;
				glBegin(GL_TRIANGLE_FAN);
				for (int i = 0; i < PNT_NUM; i++) {
					float a = std.math.PI * 2 * i / PNT_NUM;
					glVertex3f(rx*std.math.sin(a), ry*std.math.cos(a), 0);
				}
				glEnd();

				glPopMatrix();
			}

			break;
		}
		default: {
			super.draw();
		}
		}
	}

	void dyingRokFactor(int dyingFactor) {
		dyingRokFactor_ = dyingFactor;
	}
	int dyingRokFactor() {
		return dyingRokFactor_;
	}

	bool isColl(float x, float y) {
		switch (type_) {
		case (Type.SHOT_CORE):
			if (std.math.fabs(x_ - x) < 30 &&
				std.math.fabs(y_ - y) < 30) {
				if (distance(x_, y_, x, y) < 30) {
					return true;
				}
			}
			return false;

		case (Type.ENEMY_BALLS):
			if (std.math.fabs(x_ - x) < 60 &&
				std.math.fabs(y_ - y) < 60) {
				if (distance(x_, y_, x, y) < 60) {
					return true;
				}
			}
			return false;

		case (Type.ENEMY_WALL):
		case (Type.ENEMY_BLOCK):
		case (Type.ENEMY_POINT):
			if (std.math.fabs(x_ - x) < 60 &&
				std.math.fabs(y_ - y) < 20) {
				return true;
			}
			return false;
		}
		
	}

	void vanish() {
		life_ = -10000;
		score_ = 0;
	}

	void destroy() {
		static int prevTurn = 0;

		if (type_ < 0) return;

		life_ = 0;

		int t = command_.doGetGameTurn();
		if (prevTurn != t) {
			Sound.obj.playChunk(Se.SMALL_ENEMY_DESTROY);
			prevTurn = t;
		}

		command_.quit();

		dyingCnt_ = 30;
		type_ = (Type)(-type_);
	}

private:
	float score_;
	int dyingCnt_;
	int dyingRokFactor_;
}
