private import std.string;
private import std.math;

private import lua;
private import lualib;
private import lauxlib;

private import log;
private import charactor;
private import game;
private import type;
private import shot;
private import bullet;
private import kk;
private import myio;
private import mymath;
private import bl;

private:

void throwError(lua_State lua, char[] msg) {
	Log.error(msg);
}
void throwError(lua_State lua, char[] msg, int index) {
	Log.error(msg ~ ": " ~ tostr(lua_tostring(lua, index)));
}

extern (C) {
	int luaPanic(lua_State lua) {
		Log.error("lua error" ~ std.string.toString(lua_tostring(lua, 0)));
		return 0;
	}

	Bullet nowBattery;
	Command nowCommand;

	float getDouble(lua_State lua, int i) {
		if (!lua_isnumber(lua, i)) {
			throwError(lua, "incorrect argument");
		}
		return lua_tonumber(lua, i);
	}

	int getInt(lua_State lua, int i) {
		return (int)(getDouble(lua, i));
	}

	char[] getString(lua_State lua, int i) {
		if (!lua_isstring(lua, i)) {
			throwError(lua, "incorrect argument");
		}
		return std.string.toString(lua_tostring(lua, i));
	}

	int fire(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 2) {
			throwError(lua, "fire must have 2 arguments");
		}
	}
	body {
		nowCommand.addBullet(lua, nowBattery.bl,
							 getDouble(lua, 1), getDouble(lua, 2));
		return 0;
	}

	int blClear(lua_State lua) {
		nowBattery.bl.clear();
		return 0;
	}
	int blAim(lua_State lua) {
		nowBattery.bl.angleType = BLState.AngleType.AIM;
		return 0;
	}
	int blAbs(lua_State lua) {
		nowBattery.bl.angleType = BLState.AngleType.ABS;
		return 0;
	}
	int blRel(lua_State lua) {
		nowBattery.bl.angleType = BLState.AngleType.REL;
		return 0;
	}
	int blPolarCoord(lua_State lua) {
		nowBattery.bl.polar = true;
		return 0;
	}
	int blPlaneCoord(lua_State lua) {
		nowBattery.bl.polar = false;
		return 0;
	}

	int blType(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 1) {
			throwError(lua, "blType must have 1 arguments");
		}
	}
	body {
		nowBattery.bl.type = cast(Type)getInt(lua, 1);
		return 0;
	}
	int blName(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 1) {
			throwError(lua, "blName must have 1 arguments");
		}
	}
	body {
		nowBattery.bl.name = getString(lua, 1);
		return 0;
	}
	int blId(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 1 && n != 0) {
			throwError(lua, "blId must have 0 or 1 arguments");
		}
	}
	body {
		int n = lua_gettop(lua);
		if (n == 1) {
			nowBattery.bl.id = getDouble(lua, 1);
		}
		else {
			nowBattery.bl.id = float.nan;
		}
		return 0;
	}
	int blLife(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 1) {
			throwError(lua, "blLife must have 1 arguments");
		}
	}
	body {
		nowBattery.bl.life = getDouble(lua, 1);
		return 0;
	}
	int blScore(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 1) {
			throwError(lua, "blScore must have 1 arguments");
		}
	}
	body {
		nowBattery.bl.score = getDouble(lua, 1);
		return 0;
	}

	int blOffset(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 2) {
			throwError(lua, "blOffset must have 2 arguments");
		}
	}
	body {
		nowBattery.bl.x = getDouble(lua, 1);
		nowBattery.bl.y = getDouble(lua, 2);
		return 0;
	}

	int blColor(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 3) {
			throwError(lua, "blColor must have 2 arguments");
		}
	}
	body {
		nowBattery.bl.r = getDouble(lua, 1);
		nowBattery.bl.g = getDouble(lua, 2);
		nowBattery.bl.b = getDouble(lua, 3);
		return 0;
	}

	int blHoldChild(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 1) {
			throwError(lua, "blHoldChild must have 1 arguments");
		}
	}
	body {
		nowBattery.bl.holdChild = cast(bool)getInt(lua, 1);
		return 0;
	}

	int blBoss(lua_State lua) {
		nowBattery.bl.boss = true;
		return 0;
	}

	int noticeChilden(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 1) {
			throwError(lua, "noticeChildren must have 1 arguments");
		}
	}
	body {
		Command tmpCommand = nowCommand;
		Bullet tmpBattery = nowBattery;
		Bullet[] children = nowBattery.children;
		foreach (Bullet b; children) {
			b.runFunc(getString(lua, 1));
		}
		nowCommand = tmpCommand;
		nowBattery = tmpBattery;
		return 0;
	}

	int noticeParent(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 1) {
			throwError(lua, "noticeParent must have 1 arguments");
		}
	}
	body {
		Command tmpCommand = nowCommand;
		Bullet tmpBattery = nowBattery;
		nowBattery.parent.runFunc(getString(lua, 1));
		nowCommand = tmpCommand;
		nowBattery = tmpBattery;
		return 0;
	}

	int setSpeed(lua_State lua)
	in {
		int n = lua_gettop(lua);
		if (n != 2) {
			throwError(lua, "setSpeed must have 2 arguments");
		}
	}
	body {
		nowCommand.doSetSpeed(getDouble(lua, 1), getDouble(lua, 2));
		return 0;
	}
	int changeSpeed(lua_State lua) {
		int n = lua_gettop(lua);
		if (n != 1) {
			throwError(lua, "changeSpeed must have 1 arguments");
		}
		nowCommand.doChangeSpeed(getDouble(lua, 1));
		return 0;
	}
	int changeDirectionAbs(lua_State lua) {
		int n = lua_gettop(lua);
		if (n != 1) {
			throwError(lua, "changeDirection must have 1 arguments");
		}
		nowCommand.doChangeDirectionAbs(getDouble(lua, 1));
		return 0;
	}
	int changeDirectionAim(lua_State lua) {
		int n = lua_gettop(lua);
		if (n != 1) {
			throwError(lua, "changeDirection must have 1 arguments");
		}
		nowCommand.doChangeDirectionAim(getDouble(lua, 1));
		return 0;
	}
	int vanish(lua_State lua) {
		nowCommand.doVanish();
		return 0;
	}
	int vanishAllShot(lua_State lua) {
		nowCommand.doVanishAllShot();
		return 0;
	}
	int wait(lua_State lua) {
		int n = lua_gettop(lua);
		if (n != 1) {
			throwError(lua, "wait must have 1 arguments");
		}
		nowCommand.doWait(getInt(lua, 1));
		return 0;
	}
	int getRank(lua_State lua) {
		lua_pushnumber(lua, (lua_Number)0.5);
		return 1;
	}
	int getTurn(lua_State lua) {
		lua_pushnumber(lua, (lua_Number)nowCommand.doGetTurn());
		return 1;
	}
	int getGameTurn(lua_State lua) {
		lua_pushnumber(lua, (lua_Number)nowCommand.doGetGameTurn());
		return 1;
	}
	int getX(lua_State lua) {
		lua_pushnumber(lua, (lua_Number)nowBattery.x());
		return 1;
	}
	int getY(lua_State lua) {
		lua_pushnumber(lua, (lua_Number)nowBattery.y());
		return 1;
	}
	int getSpeedX(lua_State lua) {
		lua_pushnumber(lua, (lua_Number)nowBattery.sx());
		return 1;
	}
	int getSpeedY(lua_State lua) {
		lua_pushnumber(lua, (lua_Number)nowBattery.sy());
		return 1;
	}
	int getSpeed(lua_State lua) {
		lua_pushnumber(lua, (lua_Number)nowBattery.s());
		return 1;
	}
	int getDirection(lua_State lua) {
		lua_pushnumber(lua, (lua_Number)nowBattery.a());
		return 1;
	}
	int getMaxX(lua_State lua) {
		lua_pushnumber(lua, cast(lua_Number)Game.obj.gameWidth);
		return 1;
	}
	int getMaxY(lua_State lua) {
		lua_pushnumber(lua, cast(lua_Number)Game.obj.gameHeight);
		return 1;
	}
	int getPlayerX(lua_State lua) {
		lua_pushnumber(lua, cast(lua_Number)Game.obj.player().x);
		return 1;
	}
	int getPlayerY(lua_State lua) {
		lua_pushnumber(lua, cast(lua_Number)Game.obj.player().y);
		return 1;
	}
	int getPlayerAim(lua_State lua) {
		float x = Game.obj.player().x;
		float y = Game.obj.player().y;
		float a = std.math.atan2(x - nowBattery.x,
								 y - nowBattery.y);
		lua_pushnumber(lua, cast(lua_Number)a);
		return 1;
	}
	int getLife(lua_State lua) {
		lua_pushnumber(lua, cast(lua_Number)nowBattery.life);
		return 1;
	}

	int setPos(lua_State lua) {
		int n = lua_gettop(lua);
		if (n != 2) {
			throwError(lua, "setPos must have 2 arguments");
		}
		nowBattery.x = getInt(lua, 1);
		nowBattery.y = getInt(lua, 2);
		return 0;
	}

	int nextScene(lua_State lua) {
		nowCommand.doNextScene();
		return 0;
	}

	int stageEnd(lua_State lua) {
		nowCommand.doStageEnd();
		return 0;
	}

}

public:

class LuaFile {
public:
	this(char[] file) {
		Log.info(file ~ ": lua file open");

		lua = lua_open();
		lua_atpanic(lua, cast(lua_CFunction)&luaPanic);

		lua_baselibopen(lua);
		lua_mathlibopen(lua);
		lua_tablelibopen(lua);

		int ret = luaL_loadfile(lua, std.string.toStringz(file));

		if (ret != 0) {
			if (ret == LUA_ERRSYNTAX)
				throwError(lua, file ~ ": syntax error", 0);
			else if (ret == LUA_ERRMEM)
				throwError(lua, file ~ ": memory error", 0);
			else
				throwError(lua, file ~ ": some error", 0);
		}

		lua_pcall(lua, 0, LUA_MULTRET, 0);

		lua_register(lua, "fire", &fire);
		lua_register(lua, "blClear", &blClear);
		lua_register(lua, "blType", &blType);
		lua_register(lua, "blOffset", &blOffset);
		lua_register(lua, "blName", &blName);
		lua_register(lua, "blId", &blId);
		lua_register(lua, "blAim", &blAim);
		lua_register(lua, "blAbs", &blAbs);
		lua_register(lua, "blRel", &blRel);
		lua_register(lua, "blPolarCoord", &blPolarCoord);
		lua_register(lua, "blPlaneCoord", &blPlaneCoord);
		lua_register(lua, "blLife", &blLife);
		lua_register(lua, "blScore", &blScore);
		lua_register(lua, "blColor", &blColor);
		lua_register(lua, "blHoldChild", &blHoldChild);
		lua_register(lua, "blBoss", &blBoss);

		lua_register(lua, "vanish", &vanish);
		lua_register(lua, "vanishAllShot", &vanishAllShot);
		lua_register(lua, "wait", &wait);
		lua_register(lua, "setSpeed", &setSpeed);
		lua_register(lua, "changeSpeed", &changeSpeed);
		lua_register(lua, "changeDirectionAim", &changeDirectionAim);
		lua_register(lua, "changeDirectionAbs", &changeDirectionAbs);
		lua_register(lua, "getRank", &getRank);
		lua_register(lua, "getTurn", &getTurn);
		lua_register(lua, "getGameTurn", &getGameTurn);
		lua_register(lua, "getX", &getX);
		lua_register(lua, "getY", &getY);
		lua_register(lua, "getSpeedX", &getSpeedX);
		lua_register(lua, "getSpeedY", &getSpeedY);
		lua_register(lua, "getSpeed", &getSpeed);
		lua_register(lua, "getDirection", &getDirection);
		lua_register(lua, "getPlayerX", &getPlayerX);
		lua_register(lua, "getPlayerY", &getPlayerY);
		lua_register(lua, "getPlayerAim", &getPlayerAim);
		lua_register(lua, "getMaxX", &getMaxX);
		lua_register(lua, "getMaxY", &getMaxY);
		lua_register(lua, "getLife", &getLife);
		lua_register(lua, "setPos", &setPos);
		lua_register(lua, "noticeChilden", &noticeChilden);
		lua_register(lua, "noticeParent", &noticeParent);
		lua_register(lua, "nextScene", &nextScene);
		lua_register(lua, "stageEnd", &stageEnd);
	}
	lua_State lua;
}

// ȤꤢŨ lua 椹ǡ
class Command {
public:
	static bool lua_run(lua_State lua, char* func) {
		lua_getglobal(lua, func);

		if (!lua_isfunction(lua, -1)) {
			return false;
		}

        int ret = 0;
        if (nowBattery.id == float.nan) {
			ret = lua_pcall(lua, 0, 0, 0);
        }
        else {
			lua_pushnumber(lua, cast(lua_Number)nowBattery.id);
			ret = lua_pcall(lua, 1, 0, 0);
        }

        if (ret) {
			throwError(lua, std.string.toString(func) ~
					   ": function errored", 1);
		}

		return true;
	}

	this(lua_State lua, char[] fn, Bullet battery) {
		game_ = Game.obj;
		battery_ = battery;
		func_ = std.string.toStringz(fn);
		lua_ = lua;
	}

	protected bool lua_run_(char* func) {
		nowCommand = this;
		nowBattery = battery_;

		return lua_run(lua_, func);
	}

	void init() {
		char* func = toStringz(tostr(func_) ~ "_init");
		lua_run_(func);
	}

	void quit() {
		char* func = toStringz(tostr(func_) ~ "_quit");
		lua_run_(func);
	}

	void run() {
		if (!lua_run_(func_)) {
			throwError(lua_,
					   tostr(func_) ~ ": function not found");
		}
	}
	void run(char[] func) {
		char* f = toStringz(tostr(func_) ~ "_" ~ func);
		if (!lua_run_(f)) {
			throwError(lua_,
					   tostr(f) ~ ": function not found");
		}
	}

	void addBullet(lua_State lua, BLState bl, float v1, float v2) {
		float s, a, sx, sy;
		if (bl.polar) {
			s = v1;
			a = v2;
			if (bl.angleType == BLState.AngleType.AIM) {
				a += std.math.atan2(game_.player.x - battery_.x,
									game_.player.y - battery_.y);
			}
			else if (bl.angleType == BLState.AngleType.REL) {
				a += battery_.prevAngle;
			}
			sx = std.math.sin(a) * s;
			sy = std.math.cos(a) * s;
		}
		else {
			sx = v1;
			sy = v2;
			// cannot aim
			s = length(sx, sy);
			a = angle(sx, sy);
		}
		battery_.prevAngle = a;

		float x = battery_.x() + bl.x;
		float y = battery_.y() + bl.y;

		if (bl.name == "") {
			Shot shot = new Shot(x, y, sx, sy, a, s, bl.type);
			shot.setRGB(bl.r, bl.g, bl.b);
			game_.addShot(shot);
		}
		else {
			Bullet b;
			if (bl.life == -1) {
				Bullet shot = new Bullet(x, y, sx, sy, a, s,
										 lua, bl.name, bl.type, bl.id);
				shot.setRGB(bl.r, bl.g, bl.b);
				game_.addShot(shot);
				b = shot;
			}
			else {
				Enemy shot =
					new Enemy(x, y, sx, sy, a, s,
							  lua, bl.name, bl.life, bl.type, bl.id);
				shot.setRGB(bl.r, bl.g, bl.b);
				game_.addEnemy(shot);
				b = shot;

				if (bl.score != -1) {
					shot.score = bl.score;
				}
				if (bl.boss) {
					game_.bossApproach(shot);
				}
			}

			if (bl.holdChild) {
				b.parent = battery_;
				battery_.addChild(b);
			}

		}
	}

	int doGetTurn() {
		return battery_.turn();
	}
	int doGetGameTurn() {
		return game_.turn();
	}
	void doVanish() {
		game_.vanishBullet(battery_);
	}
	void doVanishAllShot() {
		game_.vanishAllShot();
	}
	void doWait(int t) {
		battery_.wait(t);
	}

	void doSetSpeed(float sx, float sy) {
		battery_.sx = sx;
		battery_.sy = sy;
		battery_.s = length(sx, sy);
		battery_.a = angle(sx, sy);
	}
	void doChangeSpeed(float s) {
		float a = std.math.atan2(battery_.sx(), battery_.sy());
		battery_.sx = std.math.sin(a)*s;
		battery_.sy = std.math.cos(a)*s;
	}
	void doChangeDirectionAbs(float a) {
		float s = std.math.sqrt(battery_.sx()*battery_.sx()+battery_.sy()*battery_.sy());
		battery_.sx = std.math.sin(a)*s;
		float sy = std.math.cos(a)*s;
		battery_.sy = sy;
		battery_.a = a;
//		battery_.noticeChangeDirection();
	}
	void doChangeDirectionAim(float a0) {
		float a = convRelAngle(a0);

		float px = game_.player().x();
		float py = game_.player().y();
		float tx = 0;//= game_.player(Colors.opposite(color_)).x();
		float ty = 0;//= game_.player(Colors.opposite(color_)).y();
		float nowAngTarget = std.math.atan2(tx-battery_.x(), ty-battery_.y());

		float s = std.math.sqrt(battery_.sx()*battery_.sx()+battery_.sy()*battery_.sy());
		a += nowAngTarget;
		battery_.sx = std.math.sin(a)*s;
		battery_.sy = std.math.cos(a)*s;
		battery_.a = a;
//		battery_.noticeChangeDirection();
	}

	void doStageEnd() {
		game_.stageEnd();
	}

private:
	float convRelAngle(float a) {
		float r = a;
		while (r < -std.math.PI) r += std.math.PI * 2;
		while (r > std.math.PI) r -= std.math.PI * 2;
		return r;
	}

	void doNextScene() {
		game_.nextScene();
	}

protected:
	this() {}

	lua_State lua_;

	Game game_;
	Bullet battery_;

	char* func_;

}

class WaitableCommand : Command {
public:
	this(LuaFile file, Bullet battery) {
		game_ = Game.obj;
		battery_ = battery;
		lua_ = file.lua;

		func_ = "top";

		lua_getglobal(lua_, func_);
		if (!lua_isfunction(lua_, -1)) {
			throwError(lua_, "function top not found");
		}

		thread_ = lua_newthread(lua_);

		lua_settop(thread_, 0);
		lua_xmove(lua_, thread_, 1);
	}

	void init() {
		/// @todo not implemented
	}

	void quit() {
		/// @todo not implemented
	}

	void run() {
		nowCommand = this;
		nowBattery = battery_;

		int res = lua_resume(lua_, 0);

		if (res) {
			throwError(lua_, "resume errored", 1);
		}
	}

private:
	lua_State thread_;
	bit running_;

}
