private import std.string;

private import SDL;
private import opengl;
private import openglu;

private import rc;
private import log;
private import std.date;
private import std.random;
private import fpstimer;
private import scene;
private import scene_game;
private import joystickmgr;
private import padinput;
private import myio;
private import letter;
private import type;
private import sound;
private import letter;

class Kk {
private:
	static Kk obj_;
public:
	static Kk obj() {
		return obj_;
	}

public:
	this(char args[][]) {
		obj_ = this;
		Log.info("init");
		initResource(args);
		initMisc();
		initSDL();
		initGL();
		Log.info("init done");
	}

	~this() {
		Log.info("quit");
		delete sound_;
		delete joystickMgr_;
		SDL_Quit();
		Log.info("quit done");
	}

	void initResource(char args[][]) {
		Log.info("init resource");

		rc_ = new Rc();
		rc().addResourceFile();
		rc().addArgument(args);
	}

	void initMisc() {
		Log.info("init misc");

		rand_seed(getUTCtime(), 0);

		initLetter();

		timer_ = new FpsTimer(rc().getBit("timerEnable"));
		timer_.setFps(60);

		sound_ = new Sound();

		end_ = false;
	}

	void initSDL() {
		Log.info("init SDL");

		int ret = SDL_Init(SDL_INIT_VIDEO|SDL_INIT_NOPARACHUTE);
		if (ret != 0) Log.errorSDL("cannot init SDL");

		ret = SDL_InitSubSystem(SDL_INIT_JOYSTICK);
		if (ret != 0) {
			Log.warnSDL("cannot init joystick");
		}

		useAudio_ = true;
		ret = SDL_InitSubSystem(SDL_INIT_AUDIO);
		if (ret != 0) {
			Log.warnSDL("cannot init audio");
			useAudio_ = false;
		}

		SDL_VideoInfo* info = SDL_GetVideoInfo();

		int bpp = Kk.rc().getInt("screenBpp");
		char[] bppMsg;

		if (bpp == -1) {
			bpp = info.vfmt.BitsPerPixel;
			Log.info("using the \"best\" bpp: " ~ tostr(bpp));
		}
		else if (bpp == 0) {
			bpp = info.vfmt.BitsPerPixel;
			Log.info("using current bpp");
		}
		else {
			Log.info("using user set bpp: " ~ tostr(bpp));
		}

		SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );  
		SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );  
		SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
		SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

		screenWidth_ = Kk.rc().getInt("screenWidth");
		screenHeight_ = Kk.rc().getInt("screenHeight");
		int screenFull = Kk.rc().getInt("screenFull");
		int flags = SDL_OPENGL;
		if (screenFull) {
			flags |= SDL_FULLSCREEN;
		}

		Log.info("set video mode " ~
				 "size = " ~ tostr(screenWidth_) ~
				 "x" ~ tostr(screenHeight_) ~
				 ", bpp = " ~ tostr(bpp) ~
				 ", flags = " ~ tostr(flags));
		if (SDL_SetVideoMode(screenWidth_, screenHeight_,
							 bpp, flags) == null)
		{
			Log.errorSDL("cannot create screen");
		}

		info = SDL_GetVideoInfo();
		Log.info("set bpp: " ~ tostr((int)info.vfmt.BitsPerPixel));

		joystickMgr_ = new JoystickMgr();
		input_ = PadInput.generate("save/key.conf");
	}

	void initGL() {
		glClearColor(0, 0, 0, 0);
		glViewport(0, 0, screenWidth_, screenHeight_);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		float asp = cast(float)screenWidth_ / screenHeight_;
		gluPerspective(45.0, asp, 0.1, 1000);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glMatrixMode(GL_MODELVIEW);

		// from PARSEC47
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_COLOR_MATERIAL);
	}

	void run() {
		Log.info("run");

		scene_ = new SceneGame();

		timer_.start();

		while (scene_ !== null) {
			scene_.run();
			if (!timer_.shouldSkip()) {
				drawSystemInfo();
				SDL_GL_SwapBuffers();
			}
			if (scene_.isEnd()) {
				scene_ = scene_.getNext();
				timer_.start();
			}
			if (!(timer_.getTurn() % 62)) {
				char[] msg =
					"FPS: " ~ tostr(timer_.getTrueFps()) ~
					"  TURN: " ~ tostr(timer_.getTurn());
				if (scene_.objectNum() != -1) {
					msg ~=
						"  OBJS: " ~ tostr(scene_.objectNum());
				}
				msg ~= "  DROPS: " ~ tostr(timer_.getDroppedTurn());
				Log.info(msg);
			}
			timer_.wait();
			endCheck();
		}

		Log.info("run done");
	}

	void drawSystemInfo() {
		drawString("FPS", 490, 180, 12, 16, 1,1,1);
		drawNumR(timer_.getTrueFps(), 555, 205, 12, 16, 1,1,1);
		drawString("OBJS", 490, 235, 12, 16, 1,1,1);
		drawNumR(scene_.objectNum(), 555, 260, 12, 16, 1,1,1);
		drawString("TURNS", 490, 290, 12, 16, 1,1,1);
		drawNumR(timer_.getTurn(), 555, 315, 12, 16, 1,1,1);
		drawString("FRM DROPS", 490, 345, 10, 16, 1,1,1);
		drawNumR(timer_.getDroppedTurn(), 555, 370, 12, 16, 1,1,1);
		drawString("SYS DROPS", 490, 400, 10, 16, 1,1,1);
		drawNumR(timer_.getSystemDroppedTurn(), 555, 425, 12, 16, 1,1,1);
	}

	void endCheck() {
		SDL_Event ev;
		while (SDL_PollEvent(&ev)) {
			if (ev.type == SDL_QUIT || input_.getButton(5)) {
				Log.info("quit event catched");
				quit();
			}
		}
		if (end_) quit();
	}

	void quit() {
		Log.info("FPS: " ~ std.string.toString(timer_.getTrueFps()));
		scene_ = null;
	}

	void end() {
		Log.info("FPS: " ~ std.string.toString(timer_.getTrueFps()));
		scene_ = scene_.getNext();
	}

	void setEnd() {
		// @@@ temporary...
		end_ = true;
	}

public:
	static Rc rc() { return obj().rc_; }

	bit useAudio() { return useAudio_; }

	PadInput padInput() { return input_; }

	FpsTimer timer() { return timer_; }

private:
	Rc rc_;

	FpsTimer timer_;

	JoystickMgr joystickMgr_;
	PadInput input_;

	bit useAudio_;

	Scene scene_;

	int screenWidth_, screenHeight_;

	Sound sound_;

	bool end_;

}

version (linux) {
	int main(char args[][]) {
		Log.init();
		Kk c = new Kk(args);
		c.run();
//	delete c;
		Log.quit();

		return 0;
	}
}
version (Win32) {
	extern(C) int doit(char args[][]) {
		Log.init();
		Kk c = new Kk(args);
		c.run();
//	delete c;
		Log.quit();

		return 0;
	}
}
