private import std.string;

private import SDL_mixer;

private import kk;
private import rc;
private import log;

enum Se {
	PLAYER_SHOT = 0,
	PLAYER_LASER,
	PLAYER_ROK,
	PLAYER_BARRIER,
	PLAYER_BARRIER_ATACK,
	PLAYER_DEAD,
	PLAYER_SHOT_HIT,
	SMALL_ENEMY_DESTROY,
	MIDDLE_ENEMY_DESTROY,
	BOSS_DESTROY,
	SE_NUM
}

/// from PARSEC47
class Sound {
private:
	static const char[][Se.SE_NUM] seNames_ = [
		Se.PLAYER_SHOT:"sePlayerShot",
		Se.PLAYER_LASER:"sePlayerLaser",
		Se.PLAYER_ROK:"sePlayerRok",
		Se.PLAYER_BARRIER:"sePlayerBarrier",
		Se.PLAYER_BARRIER_ATACK:"sePlayerBarrierAtack",
		Se.PLAYER_DEAD:"sePlayerDead",
		Se.PLAYER_SHOT_HIT:"sePlayerShotHit",
		Se.SMALL_ENEMY_DESTROY:"seSmallEnemyDestroy",
		Se.MIDDLE_ENEMY_DESTROY:"seMiddleEnemyDestroy",
		Se.BOSS_DESTROY:"seBossDestroy"
		];

	static const int[Se.SE_NUM] seChannels_ = [
		Se.PLAYER_SHOT:0,
		Se.PLAYER_LASER:0,
		Se.PLAYER_ROK:1,
		Se.PLAYER_BARRIER:0,
		Se.PLAYER_BARRIER_ATACK:0,
		Se.PLAYER_DEAD:0,
		Se.PLAYER_SHOT_HIT:2,
		Se.SMALL_ENEMY_DESTROY:3,
		Se.MIDDLE_ENEMY_DESTROY:4,
		Se.BOSS_DESTROY:4
		];

private:
	bool noSound_ = false;
	int fadeOutSpeed_ = 1280;

	static Sound obj_;

public:
	static Sound obj() { return obj_; }

public:
	static const char[] soundsDir = "data/sounds/";
	static const char[] chunksDir = "data/sounds/";

	this() {
		obj_ = this;

		noSound_ = Kk.rc.getBit("noSound");

		if (noSound_) return;

		int audio_rate;
		Uint16 audio_format;
		int audio_channels;
		int audio_buffers;

		if (SDL_InitSubSystem(SDL_INIT_AUDIO) < 0) {
			noSound_ = 1;
			Log.warnSDL("Unable to initialize SDL_AUDIO");
			return;
		}

		audio_rate = 44100;
		audio_format = AUDIO_S16;
		audio_channels = 1;
		audio_buffers = 4096;
		if (Mix_OpenAudio(audio_rate, audio_format,
						  audio_channels, audio_buffers) < 0) {
			noSound_ = 1;
			Log.warnSDL("Couldn't open audio");
			return;
		}
		Mix_QuerySpec(&audio_rate, &audio_format, &audio_channels);

		loadChunks();
	}

	void close() {
		if (noSound_) return;
		if (Mix_PlayingMusic()) {
			Mix_HaltMusic();
		}
		Mix_CloseAudio();
	}

	// Free a music or a chunk.
	~this() {
		if (music_) {
			stopMusic();
			Mix_FreeMusic(music_);
			music_ = null;
		}
		foreach (Mix_Chunk* c; chunk_) {
			if (c) {
				Mix_FreeChunk(c);
				c = null;
			}
		}
		for (int i = 0; i < MIX_CHANNELS; i++) {
			haltChunk(i);
		}
	}

	// Load a sound or a chunk.

	void loadMusic(char[] name) {
		if (noSound_) return;

		if (music_) {
			stopMusic();
			Mix_FreeMusic(music_);
			music_ = null;
		}

		char[] fileName = soundsDir ~ name;
		music_ = Mix_LoadMUS(toStringz(fileName));
		if (!music_) {
//			noSound_ = true;
			Log.warnSDL("Couldn't load: " ~ fileName);
		}
	}
  
	// Play/Stop the music/chunk.

	void playMusic() {
		if (noSound_) return;
		if (music_) {
			Mix_PlayMusic(music_, -1);
		}
	}

	void fadeMusic() {
		if (noSound_) return;
		Mix_FadeOutMusic(fadeOutSpeed_);
	}

	void stopMusic() {
		if (noSound_) return;
		if (Mix_PlayingMusic() ) {
			Mix_HaltMusic();
		}
	}

	void playChunk(Se id) {
		if (noSound_) return;
		if (chunk_[id]) {
			Mix_PlayChannel(seChannels_[id], chunk_[id], 0);
		}
	}

	void haltChunk(int ch) {
		if (noSound_) return;
		Mix_HaltChannel(ch);
	}
	void haltChunk(Se id) {
		if (noSound_) return;
		Mix_HaltChannel(seChannels_[id]);
	}

private:
	void loadChunks() {
		if (noSound_) return;

		for (int i = 0; i < Se.SE_NUM; i++) {
			char[] seName = seNames_[i];
			char[] seFileName = chunksDir ~ Kk.rc.get(seName);
			chunk_[i] = Mix_LoadWAV(toStringz(seFileName));
			if (!chunk_[i]) {
				Log.warnSDL("Couldn't load: " ~ seFileName);
			}
		}
	}

private:
	Mix_Music* music_;
	Mix_Chunk*[Se.SE_NUM] chunk_;

}

/**
@memo
usage image ...

Sound.obj.playMusic(1);
Sound.obj.fadeMusic();
Sound.obj.haltMusic();
Sound.obj.playChunk(PLAYER_SHOT_A);
Sound.obj.haltChunk(PLAYER_SHOT_A);

*/

